import unittest

from specialsauce.sources.trainingpeaks import ewma_days, sma


class TestEwmaDays(unittest.TestCase):
  def test_stable(self):
    init = 100.0
    result = ewma_days(
      [init if i != 0 else 0 for i in range(5)],
      7,
      init=init
    )
    for el in result:
      self.assertAlmostEqual(el, init)


class TestSma(unittest.TestCase):
  def test_sma(self):
    # x = pd.Series([10 for i in range(60)])
    # x.rolling(30).mean()  # nan until 29

    # x.index = [pd.to_datetime(2*i, unit='s') for i in range(len(x))]
    # x = pd.Series([i for i in range(60)])
    # x.index = [pd.to_datetime(2*i, unit='s') for i in range(len(x))]
    # x.rolling('30s').mean()

    pass


import os

import pandas as pd


def load_ngp_gap():
    """Load and return the NGP / GAP dataset.
    
    Inspired by sklearn's `load_iris()`.
    """
    module_path = os.path.dirname(__file__)
    ngp_gap_csv_filename = os.path.join(module_path, 'data', 'ngp_gap.csv')
    df_adjusted_pace_str = pd.read_csv(ngp_gap_csv_filename, index_col='Grade')
    df_adjusted_pace_td = df_adjusted_pace_str  \
        .applymap(lambda x: f'00:{x}' if not pd.isnull(x) else x)  \
        .applymap(pd.to_timedelta) 
        # .apply(lambda series: series.dt.total_seconds())

    return df_adjusted_pace_td



import pandas as pd


def ewma(x_arr, time_arr, alpha, init=0.0):
  """Exponentially-weighted moving average.
  
  Behaves like O2 consumption or training load - takes a while to reach
  steady-state when starting out. This is different than the pandas EWMA
  implementation, which keeps the first value of x-arr as the first
  value of the average.
  
  Args:
    x_arr (list(float)): A series of values to calculate a EWMA of.
    time_arr (list(datetime.timedelta, datetime.datetime, or datetime.date)):
      Times corresponding to each value of x_arr.
    alpha (float): Decay coefficient for the EWMA. Must be 0 <= alpha <= 1.
      The smaller the alpha, the more important old values are to the
      average, or in other words longer the average's memory.
  """
  time_arr = pd.Series(time_arr)

  # Convert to timedelta with first value at 0.
  timedelta_array = time_arr - time_arr[0]

  # Convert to decimal days
  delta_day_decimal_array = [d.total_seconds() / 86400 for d in timedelta_array]

  # Initialize the moving average
  x_avg_arr = [init + x_arr[0] * alpha]

  # Recursively calculate the average at each time step
  for i in range(1, len(x_arr)):
    x_avg_arr.append(
      x_avg_arr[i-1] * (1 - alpha) ** (delta_day_decimal_array[i] - delta_day_decimal_array[i-1])
      + x_arr[i] * alpha
    )

  return x_avg_arr


def ewma_pandas(x_series, half_life, time_series=None):
  """Exponentially-weighted moving average.
  
  Behaves like O2 consumption - takes a while to reach steady-state
  when starting out.

  Args:
    x_series (pandas.Series): Values to make a EWMA of.
    half_life (int, str, or pandas.timedelta): half-life of the EWMA.
      If int, and time_series is provided, assumed to be integer seconds.
    time_series (pandas.Series): integer seconds from the start of the
      activity. If present, these will be used as coordinates over which
      we take the moving average. Default None.
  """

  # (no longer necessary) Calculate alpha from half-life.
  # alpha = 1 - math.exp(-math.log(2) / half_life)
  # alpha = 1 - math.exp(-1 / tau)

  x_series_pad = pd.concat([pd.Series([0.0]), x_series])

  if time_series is None:
    ewm_pad = x_series_pad.ewm(
      # alpha=(1 - math.exp(-1 / tau)),
      # halflife=(-1.0 * math.log(0.5) * tau),
      halflife=half_life,
      adjust=False,
      ignore_na=True,
    ).mean()

    return ewm_pad[1:]
  

def ewma_halflife(x_series, half_life, time_series=None):
  """Exponentially-weighted moving average.
  
  Behaves like O2 consumption - takes a while to reach steady-state
  when starting out.

  Args:
    x_series (pandas.Series): Values to make a EWMA of.
    half_life (int, str, or pandas.timedelta): half-life of the EWMA.
      If int, and time_series is provided, assumed to be integer seconds.
    time_series (pandas.Series): integer seconds from the start of the
      activity. If present, these will be used as coordinates over which
      we take the moving average. Default None.
  """

  # (no longer necessary) Calculate alpha from half-life.
  # alpha = 1 - math.exp(-math.log(2) / half_life)
  # alpha = 1 - math.exp(-1 / tau)

  x_series_pad = pd.concat([pd.Series([0.0]), x_series])

  if time_series is None:
    ewm_pad = x_series_pad.ewm(
      # alpha=(1 - math.exp(-1 / tau)),
      # halflife=(-1.0 * math.log(0.5) * tau),
      halflife=half_life,
      adjust=False,
      ignore_na=True,
    ).mean()

    return ewm_pad[1:]
  else:
    # Assume we are working with seconds.
    if isinstance(half_life, (int, float)):
      half_life = pd.to_timedelta(half_life, unit='s')
    elif isinstance(half_life, str):
      half_life = pd.to_timedelta(half_life)

    # Initialize the moving average so it takes off from 0 and tends
    # toward steady-state.
    num_padding = int(half_life.seconds * 40)
    x_series_pad = pd.Series(
      [0.0 for i in range(num_padding)] + x_series.to_list()
    )
    time_series = time_series - time_series[0]
    time_series_pad = pd.Series(
      [i for i in range(num_padding)] + (time_series + num_padding).to_list(),
    ).apply(pd.to_datetime, unit='s')

    # time_series_pad = pd.concat([
    #   pd.Series([0], dtype='datetime64[s]'),
    #   (time_series + 1).apply(pd.to_datetime, unit='s')
    # ])

    ewm_pad = x_series_pad.ewm(
      # alpha=(1 - math.exp(-1 / tau)),
      # halflife=(-1.0 * math.log(0.5) * tau),
      halflife=half_life,
      times=time_series_pad,
      adjust=False,
      ignore_na=True,
    ).mean()

    ewm = ewm_pad[num_padding:]
    ewm.index = x_series.index

    return ewm


import unittest

import pandas as pd

from specialsauce.sources.minetti import cost_of_running, cost_of_walking
from specialsauce.sources.strava import gap_speed_factor
from specialsauce.sources.trainingpeaks import ngp_speed_factor


class FactorTestMixin:
  func = None
  range_min = None
  range_max = None

  @classmethod
  def setUpClass(cls):
    cls.func = staticmethod(cls.func)

  def test_below_range(self):
    self.assertEqual(
      self.func(self.range_min),
      self.func(self.range_min - 0.01)
    )

  def test_above_range(self):
    self.assertEqual(
      self.func(self.range_max),
      self.func(self.range_max + 0.01)
    )

  def test_percent_grade_input(self):
    self.assertEqual(
      self.func(self.range_max),
      self.func(self.range_max * 100)
    )

    self.assertEqual(
      self.func(self.range_min),
      self.func(self.range_min * 100)
    )

  def test_below_range_series(self):
    result = self.func(pd.Series(
      [self.range_min - 0.01] * 50
    ))
    self.assertEqual(
      result.min(),
      self.func(self.range_min)
    )

  def test_above_range_series(self):
    result = self.func(pd.Series(
      [self.range_max + 0.01] * 50
    ))
    self.assertEqual(
      result.max(),
      self.func(self.range_max)
    )


class TestMinettiRun(FactorTestMixin, unittest.TestCase):
  func = cost_of_running
  range_min = -0.45
  range_max = 0.45


class TestMinettiWalk(FactorTestMixin, unittest.TestCase):
  func = cost_of_walking
  range_min = -0.45
  range_max = 0.45


class TestNgp(FactorTestMixin, unittest.TestCase):
  func = ngp_speed_factor
  range_min = -0.25
  range_max = 0.30


class TestGap(FactorTestMixin, unittest.TestCase):
  func = gap_speed_factor
  range_min = -0.45
  range_max = 0.45



import numpy as np
import pandas as pd
from scipy.interpolate import interp1d


def ngp_speed_factor(decimal_grade):
  """Calculate TrainingPeaks' NGP pace-factor as a function of percent grade.

  The factor will be greater than 1.0 if NGP is faster than horizontal speed,
  and less than 1.0 if NGP is slower than horizontal speed.

  Grade is constrained to (-25%, 30%), because that was the range of grades
  that yielded reasonable paces in my investigation:
    * 0.45 adjusted 8:00 pace to 0:17 pace.
    * The adjustment factor jumped wildly from -26% to -27% and from
      31% to 32%.

  Rather than extrapolating outside of that range, I chose to cap the 
  adjustment factor.

  TrainingPeaks does not provide any info about how they derived their
  grade adjustment factor, but the shape of the curve sure looks a lot
  like Strava's GAP.

  Args:
    grade (float): Decimal grade to evaluate the NGP speed-factor.
  Returns:
    float: Factor that converts speed to NGP adjusted speed.
  """
  # Constrain decimal grade to the range of the equation's validity
  decimal_grade = np.clip(decimal_grade, -0.25, 0.3)

  adjustment_factors = {
    45: 28.235,
    32: 20.0,
    31: 3.556,
    30: 3.429,
    25: 2.874,
    20: 2.365,
    15: 1.905,
    10: 1.514,
    8: 1.383,
    6: 1.266,
    4: 1.162,
    2: 1.074,
    0: 1.0,
    -2: 0.941,
    -4: 0.897,
    -6: 0.87,
    -8: 0.856,
    -10: 0.857,
    -15: 0.921,
    -20: 1.067,
    -25: 1.28,
    -26: 1.33,
    -27: 2.051,
    -30: 2.297,
    -45: 3.934
  }

  interp_fn = interp1d(
    list(adjustment_factors.keys()),
    list(adjustment_factors.values())
  )

  return interp_fn(decimal_grade * 100)


def normalize(series):
  """Calculates the TrainingPeaks norm of a series of data.

  This works like a modified version of an average, based on principles
  of human physiology.

  First, a 30-second rolling average is applied to the series. 
  Since this norm is meant to replicate TrainingPeaks NP/NGP, I follow
  their lead in throwing out the first `window_len` of data 
  (30 sec in their case).

  Note: I'm not sure if this means they insert 0 values for the first
  30 seconds, or if they take an average of points after 30 seconds.
  My EWMA function (below) slowly increments until it gets to a
  steady-state - so it is biased to be low for short-duration bouts.

  Then, the L4 norm is applied to the smoothed series. This norm 
  emphasizes higher values by raising all values to the 4th power, 
  averaging them, then taking the 4th root.
  
  Emphasizing higher values is based in the idea that best-fit curves 
  between VO2 (oxygen consumption) and blood lactate tend to follow a 
  4th-order relationship.

  References:
    https://help.trainingpeaks.com/hc/en-us/articles/204071804-Normalized-Power

  """
  series_rolling = sma(series, 30)
  return l4_norm(series_rolling[29:])


def sma(x_series, window_len, time_series=None):
  """Simple moving average, as implemented by TrainingPeaks."""

  if time_series is None:
    sma = x_series.rolling(window_len).mean()
  else:
    # Assume we are working with seconds.
    if isinstance(window_len, (int, float)):
      window_len = pd.to_timedelta(window_len, unit='s')
    elif isinstance(window_len, str):
      window_len = pd.to_timedelta(window_len)

    x_series.index = time_series.map(lambda x: pd.to_datetime(x, unit='s'))

    sma = x_series.rolling(window_len).mean()

    sma.index = time_series.index

  return sma


def l4_norm(series):
  return (series ** 4).mean() ** 0.25


def training_stress_score(ngp, ftp, duration_sec):
  """TSS"""
  return 100 * (ngp / ftp) ** 2 * (duration_sec / 3600)


def acute_training_load(tss_array, init=0.0):
  """ATL

  Args:
    tss_arr (list-like): array of Training Stress Score values, assumed
      to occur exactly [0, 1, 2, ...] days from now.
  """
  return ewma_days(tss_array, 7, init)


def chronic_training_load(tss_array, init=0.0):
  """CTL

  Args:
    tss_arr (list-like): array of Training Stress Score values, assumed
      to occur exactly [0, 1, 2, ...] days from now.
  """
  return ewma_days(tss_array, 42, init)


def ewma_days(x_array, n_days, init=0.0):
  alpha = 1 / n_days
  
  return [
    init * (1 - alpha) ** i + sum(
      x_array[i - n] * alpha * (1 - alpha) ** n 
      for n in range(i + 1)) 
    for i in range(len(x_array))
  ]


def training_status(training_stress_balance):
  if training_stress_balance > 25:
    return 'Likely losing fitness quickly'
  elif training_stress_balance > 15:
    return 'Extensively recovered and likely losing fitness'
  elif training_stress_balance > -10:
    return 'Maintenance'
  elif training_stress_balance > -30:
    return 'Productive'
  else:
    return 'Overreaching'



import unittest

import pandas as pd

from specialsauce.datasets import load_ngp_gap


class TestDatasets(unittest.TestCase):
    def test_load_ngp_gap(self):
        data = load_ngp_gap()
        self.assertIsInstance(data, pd.DataFrame)
        self.assertTrue(all(
            [pd.api.types.is_timedelta64_dtype(dtype) 
             for dtype in data.dtypes]
        ))


import unittest

import datetime
import math
import os

import numpy as np
import pandas as pd

import specialsauce
from specialsauce import util as putil

BASE_DIR = os.path.abspath(os.path.dirname(__file__))


@unittest.skip('I do not have the moving average funcs cleaned up yet')
class TestMovingAverage(unittest.TestCase):
  def test_ewma(self):
    n = 180
    s = pd.Series([10.0 for i in range(n)])
    # s.iloc[2:5] = [np.nan] * 3
    
    tau = 100
    t = pd.Series([t for t in range(len(s))])
    s_theory = 10.0 * (1 - np.exp(-(t+1) / tau))

    halflife = math.log(2) * tau
    
    # print(ewma(s, halflife))
    self.assertAlmostEqual(
      (putil.ewma(s, halflife) - s_theory).max(),
      0
    )

    # print(putil.ewma(s, f'{halflife}s', time_series=t))
    # print(ewma(s, halflife, time_series=t))
    # print(ewma(s, f'{halflife}s', time_series=t) - s_theory)
    self.assertAlmostEqual(
      (putil.ewma(s, f'{halflife}s', time_series=t) - s_theory).max(),
      0
    )

    times = pd.Series([30 + 2 * t for t in range(len(s))])
    s_theory = pd.Series([10.0 * (1 - np.exp(-(2 * t + 1) / tau)) for t in range(len(times))])
    # print(putil.ewma(s, f'{halflife}s', time_series=times))
    self.assertAlmostEqual(
      (putil.ewma(s, f'{halflife}s', time_series=times) - s_theory).max(),
      0
    )




import math

import numpy as np
import pandas as pd

from specialsauce import util
from specialsauce.sources import minetti


def power_met_ss(speed, grade=0.0):
  return speed * minetti.cost_of_running(grade)


def power_met_ss(speed_series, grade_series=None):
  """Calculate steady-state metabolic power in the moderate domain.

  For more info, see `heartandsole_local/heartandsole/powerutils.py`.
  """
  if grade_series is None:
    grade_series = pd.Series([0 for i in range(len(speed_series))])

  c_i_series = grade_series.apply(minetti.cost_of_running)

  # Instantaneous running power (W/kg) is simply cost of running 
  # (J/kg/m) multiplied by speed (m/s).
  power_series = c_i_series * speed_series
  
  # Updated to account for the fact that the horizontal speed is
  # measured, but cost of running relates to the distance along the
  # incline.
  power_series = power_series / np.cos(np.arctan(grade_series))

  return power_series


def power_met(speed_series, grade_series=None, time_series=None, tau=20):
  """Calculate metabolic power in the moderate domain as a time series."""
  if time_series is None:
    time_series = pd.Series([i for i in range(len(speed_series))])

  # Calculate the theoretical steady-state power associated with the
  # speed and grade value at each timestep.
  power_met_inst = power_met_ss(speed_series, grade_series=grade_series)

  halflife = tau * math.log(2)

  # Even if we have uniform 1-second samples, I would still need this
  # util function. It makes the average start at 0 and trend up, rather
  # than letting the first value have all the weight.

  return util.ewma_halflife(
    power_met_inst,
    halflife,
    time_series=time_series,
  )


def run_cost(grade=0.0):
  """Calculates the metabolic cost of running.

  See the documentation for powerutils.o2_power_ss for information
  on the scientific basis for this calculation.

  Args:
    speed (float): Running speed in meters per second. 
    grade (float): Decimal grade, i.e. 45% = 0.45.

  Returns:
    float: Cost of running on an inclined treadmill, in Joules/kg/m,
      with distance measured along the incline slope.
  """
  c_i = minetti.cost_of_running(grade)

  return c_i


def run_power(speed, grade=0.0):
  return run_cost(speed, grade=grade) * speed / math.cos(math.atan(grade))
